/* References
----------
T. Fujiwara, J.-K. Chou, Shilpika, P. Xu, L. Ren, K.-L. Ma. An Incremental
Dimensionality Reduction Method for Visualizing Streaming Multidimensional
Data, TVCG, 2020
*/

#include "inc_pca.hpp"

#include <iostream>

IncPCA::IncPCA(Eigen::Index const nComponents, float const forgettingFactor)
    : nComponents(nComponents), forgettingFactor(forgettingFactor) {
  initialize();
}

void IncPCA::initialize() {
  components_.resize(0, 0);
  nSamplesSeen_ = 0;
  mean_ = 0.0f;
  var_ = 0.0f;
  singularValues_.resize(0, 0);
  explainedVariance_.resize(0, 0);
  explainedVarianceRatio_.resize(0, 0);
  noiseVariance_ = 0.0f;
}

void IncPCA::partialFit(Eigen::MatrixXf X) {
  auto nSamples = X.rows();
  auto nFeatures = X.cols();

  // sanity check
  if (nComponents < 1) {
    nComponents_ = components_.rows() < 1 ? std::min(nSamples, nFeatures)
                                          : components_.rows();
  } else if (nComponents > nFeatures) {
    std::cerr << "nComponents must be smaller than or equal to nFeatures"
              << std::endl;
    return;
  } else if (nComponents > nSamples) {
    std::cerr << "nComponents must be smaller than or equal to nSamples"
              << std::endl;
  } else {
    nComponents_ = nComponents;
  }

  if (components_.rows() >= 1 && components_.rows() != nComponents_) {
    std::cerr << "size error";
  }

  // Initialize mean_ and var_
  if (nSamplesSeen_ == 0) {
    mean_ = Eigen::ArrayXf::Zero(nFeatures);
    var_ = Eigen::ArrayXf::Zero(nFeatures);
  }

  // Update stats
  auto colMean_colVar_nTotalSamples =
      incrementalMeanAndVar(X, forgettingFactor, mean_, var_, nSamplesSeen_);
  Eigen::RowVectorXf colMean = std::get<0>(colMean_colVar_nTotalSamples);
  Eigen::RowVectorXf colVar = std::get<1>(colMean_colVar_nTotalSamples);
  auto nTotalSamples = std::get<2>(colMean_colVar_nTotalSamples);

  // Whitening
  if (nSamplesSeen_ == 0) {
    // If it is the first step, simply whiten X
    X = X.rowwise() - colMean;
  } else {
    Eigen::RowVectorXf colBatchMean = X.colwise().mean();
    // If X has one row, we should not subtract mean (everything col becomes 0)
    if (X.rows() > 1) {
      X = X.rowwise() - colBatchMean;
    }
    Eigen::RowVectorXf castedMean_ = mean_;
    // Build matrix of combined previous basis and new data
    Eigen::RowVectorXf meanCorrection =
        std::sqrt((nSamplesSeen_ * nSamples) / nTotalSamples) *
        (castedMean_ - colBatchMean);

    singularValues_ = singularValues_.array() * forgettingFactor;
    Eigen::Map<Eigen::RowVectorXf> rvecSingularValues_(singularValues_.data(),
                                                       singularValues_.size());

    Eigen::MatrixXf sc = rvecSingularValues_ * components_;
    // vertical stack of result
    Eigen::MatrixXf tmpX(sc.rows() + X.rows() + meanCorrection.rows(),
                         sc.cols());
    tmpX << rvecSingularValues_ * components_, X, meanCorrection;
    X = tmpX;
  }

  // SVD decomposition
  Eigen::BDCSVD<Eigen::MatrixXf> svd(X,
                                     Eigen::ComputeThinU | Eigen::ComputeThinV);
  Eigen::MatrixXf U = svd.matrixU();
  Eigen::MatrixXf V =
      svd.matrixV().transpose(); // Eigen and numpy linalg svg has different
                                 // specification for return V
  Eigen::VectorXf S = svd.singularValues();

  svdFlip(U, V);

  Eigen::ArrayXf explainedVariance =
      pow(S.array(), 2.0f) / float(nTotalSamples - 1);
  Eigen::ArrayXf explainedVarianceRatio =
      pow(S.array(), 2.0f) / (colVar.array() * float(nTotalSamples)).sum();
  nSamplesSeen_ = Eigen::Index(forgettingFactor * nSamplesSeen_ +
                               (nTotalSamples - nSamplesSeen_));

  auto nRowsToTake = std::min(nComponents_, V.rows());
  components_ = V.block(0, 0, nRowsToTake, V.cols());
  singularValues_ = S.block(0, 0, nRowsToTake, S.cols());

  mean_ = colMean;
  var_ = colVar;
  explainedVariance_ =
      explainedVariance.block(0, 0, nRowsToTake, explainedVariance.cols());
  explainedVarianceRatio_ = explainedVarianceRatio.block(
      0, 0, nRowsToTake, explainedVarianceRatio.cols());
  if (nRowsToTake < nFeatures) {
    if (explainedVariance.rows() > nRowsToTake) {
      noiseVariance_ =
          explainedVariance
              .block(nRowsToTake, 0, explainedVariance.rows() - nRowsToTake,
                     explainedVariance.cols())
              .array()
              .mean();
    }
  } else {
    noiseVariance_ = 0.0f;
  }
}

std::tuple<Eigen::ArrayXf, Eigen::ArrayXf, Eigen::Index>
IncPCA::incrementalMeanAndVar(Eigen::MatrixXf const &X, float forgettingFactor,
                              Eigen::ArrayXf const &lastMean,
                              Eigen::ArrayXf const &lastVariance,
                              Eigen::Index const lastSampleCount) {
  /* References
  ----------
  T. Chan, G. Golub, R. LeVeque. Algorithms for computing the sample
      variance: recommendations, The American Statistician, Vol. 37, No. 3,
      pp. 242-247
  */

  Eigen::ArrayXf lastSum = forgettingFactor * lastMean * float(lastSampleCount);
  Eigen::ArrayXf newSum = X.colwise().sum();
  auto newSampleCount = X.rows();
  auto updatedSampleCount =
      forgettingFactor * float(lastSampleCount) + float(newSampleCount);
  Eigen::ArrayXf updatedMean = (newSum + lastSum) / float(updatedSampleCount);

  Eigen::ArrayXf updatedVariance;
  Eigen::ArrayXf updatedUnnormalizedVariance;
  if (lastVariance.rows() != 0) {
    // colwise variace * newSampleCount
    Eigen::ArrayXf newUnnormalizedVariance =
        (pow(X.array(), 2.0f).colwise().mean() -
         pow(X.colwise().mean().array(), 2.0f)) *
        float(newSampleCount);

    if (lastSampleCount == 0) { // Avoid division by 0
      updatedUnnormalizedVariance = newUnnormalizedVariance;
    } else {
      auto lastOverNewCount = float(lastSampleCount) / float(newSampleCount);
      Eigen::ArrayXf lastUnnormalizedVariance =
          lastVariance * float(lastSampleCount);

      updatedUnnormalizedVariance =
          lastUnnormalizedVariance + newUnnormalizedVariance +
          lastOverNewCount / float(updatedSampleCount) *
              pow((lastSum / lastOverNewCount - newSum), 2.0f);
    }
    updatedVariance = updatedUnnormalizedVariance / float(updatedSampleCount);
  }

  return std::make_tuple(updatedMean, updatedVariance,
                         lastSampleCount + newSampleCount);
}

void IncPCA::svdFlip(Eigen::MatrixXf &U, Eigen::MatrixXf &V) {
  auto n = U.cols();
  for (Eigen::Index i = 0, m = V.rows(); i < n; ++i) {
    Eigen::MatrixXf::Index maxIndex = 0;
    // Eigen maxCoeff doesn't work well for single cell matrix
    auto maxAbsPos =
        m == 1 ? 0 : Eigen::Index(V.row(i).array().abs().maxCoeff(&maxIndex));

    auto value = V(i, maxAbsPos);
    auto sign = (value > 0.0f) - (value < 0.0f);
    if (n > i)
      U.col(i) *= sign;
    V.row(i) *= sign;
  }
}

Eigen::MatrixXf IncPCA::transform(Eigen::MatrixXf const &X) {
  Eigen::MatrixXf result;
  if (components_.rows() > 0 && X.cols() == components_.cols()) {
    result = X.rowwise() - Eigen::RowVectorXf(mean_);
    result *= components_.transpose();
  } else {
    std::cerr << "Matrix cols and components cols have different sizes."
              << std::endl;
  }

  return result;
}

Eigen::MatrixXf IncPCA::getComponents() { return components_.transpose(); }

Eigen::MatrixXf IncPCA::getLoadings() {
  Eigen::MatrixXf result(components_.rows(), components_.cols());
  if (components_.rows() > 0) {
    Eigen::MatrixXf sqrtLambda = explainedVarianceRatio_.array().sqrt();
    for (Eigen::Index row = 0, m = components_.rows(); row < m; ++row) {
      for (Eigen::Index col = 0, n = components_.cols(); col < n; ++col) {
        result(row, col) = components_(row, col) * sqrtLambda(row);
      }
    }
  }
  return result;
}

float IncPCA::getUncertV(unsigned int const nObtainedFeatures) {
  auto result = 0.0f;
  if (components_.rows() > 0) {
    auto k = std::min(components_.rows(), explainedVarianceRatio_.rows());
    auto sumR = explainedVarianceRatio_.block(0, 0, k, 1).array().sum();
    auto sumCoveredVal = 0.0f;
    auto nFeatures = components_.cols();

    for (Eigen::Index i = 0; i < k; ++i) {
      Eigen::MatrixXf w = components_.block(i, 0, 1, nFeatures).array() *
                          std::sqrt(singularValues_(i, 0));

      auto a = w.block(0, 0, 1, nObtainedFeatures).array().abs().sum();
      auto b = w.array().abs().sum();
      b = std::max(b, 0.000000000000001f); // to avoid zero div
      sumCoveredVal += a / b;
    }
    sumR = std::max(sumR, 0.000000000000001f); // to avoid zero div
    result = sumCoveredVal / float(k);
  }
  return 1.0f - result;
}

Eigen::MatrixX2f IncPCA::geomTrans2D(Eigen::MatrixX2f const &pointsFrom,
                                     Eigen::MatrixX2f const &pointsTo) {
  auto n = std::min(pointsFrom.rows(), pointsTo.rows());
  auto m = pointsTo.rows() - pointsFrom.rows();

  if (m < 0) {
    std::cerr << "geomTrans2D: does not support the case that rows of pointsTo "
                 "is smaller than rows of pointsFrom"
              << std::endl;
  }

  Eigen::MatrixX2f processedPointsFrom = pointsFrom.topRows(n);
  Eigen::MatrixX2f processedPointsTo = pointsTo.topRows(n);
  // translation
  Eigen::RowVector2f meanPointsFrom = processedPointsFrom.colwise().mean();
  Eigen::RowVector2f meanPointsTo = processedPointsTo.colwise().mean();
  processedPointsFrom = processedPointsFrom.rowwise() - meanPointsFrom;
  processedPointsTo = processedPointsTo.rowwise() - meanPointsTo;

  // uniform scaling
  float scalePointsFrom = processedPointsFrom.colwise().norm().sum();
  float scalePointsTo = processedPointsTo.colwise().norm().sum();
  scalePointsFrom /= float(n);
  scalePointsTo /= float(n);
  scalePointsFrom = std::sqrt(scalePointsFrom);
  scalePointsTo = std::sqrt(scalePointsTo);
  processedPointsFrom /= scalePointsFrom;
  processedPointsTo /= scalePointsTo;

  // rotation
  Eigen::BDCSVD<Eigen::MatrixXf> svd(processedPointsFrom.transpose() *
                                         processedPointsTo,
                                     Eigen::ComputeThinU | Eigen::ComputeThinV);
  Eigen::MatrixXf U = svd.matrixU();
  Eigen::MatrixXf V = svd.matrixV(); // before transposed unlike numpy svd
  Eigen::VectorXf S = svd.singularValues();
  Eigen::MatrixXf R = V * U.transpose();

  // apply tranformation
  Eigen::MatrixX2f transformedPointsTo(pointsTo.rows(), 2);
  transformedPointsTo.topRows(n) = scalePointsFrom * processedPointsTo * R;

  // for new points
  Eigen::MatrixX2f processedNewPointsTo = pointsTo.bottomRows(m);
  processedNewPointsTo = processedNewPointsTo.rowwise() - meanPointsTo;
  processedNewPointsTo /= scalePointsTo;
  processedNewPointsTo = scalePointsFrom * processedNewPointsTo * R;

  // aggregate result and move center
  transformedPointsTo.bottomRows(m) = processedNewPointsTo;
  transformedPointsTo = transformedPointsTo.rowwise() + meanPointsFrom;

  return transformedPointsTo;
}

void IncPCA::updateWithAdadelta(Eigen::Vector3f &theta,
                                Eigen::Vector3f &meanSqGrad,
                                Eigen::Vector3f &meanSqDTheta,
                                Eigen::Vector3f const &grad, float const rho,
                                float const epsilon) {
  meanSqGrad = rho * meanSqGrad.array() + (1.0f - rho) * grad.array().square();
  Eigen::Vector3f const dTheta = -((meanSqDTheta.array() + epsilon).sqrt() /
                                   (meanSqGrad.array() + epsilon).sqrt()) *
                                 grad.array();
  meanSqDTheta =
      rho * meanSqDTheta.array() + (1.0f - rho) * dTheta.array().square();

  theta = theta.array() + dTheta.array();
}

std::pair<Eigen::RowVector2f, float>
IncPCA::posEst(Eigen::RowVector2f const &posInPointsFrom,
               Eigen::MatrixX2f const &pointsFrom,
               Eigen::MatrixX2f const &pointsTo) {
  // calc distances
  Eigen::VectorXf dists =
      (pointsFrom.rowwise() - posInPointsFrom).rowwise().norm();

  // initialize position with centroid of pointsTo
  Eigen::RowVector2f initPos = pointsTo.colwise().mean();

  Eigen::Vector3f theta;
  theta << initPos(0), initPos(1), 1.0f; // 1.0f is initial lambda
  Eigen::Vector3f estimatedTheta = optPosEst(theta, dists, pointsTo);
  float uncertU = calcUncertU(theta, dists, pointsTo);

  Eigen::RowVector2f pos;
  pos << estimatedTheta(0), estimatedTheta(1);

  return {pos, uncertU};
}

Eigen::Vector3f IncPCA::optPosEst(Eigen::Vector3f const &initTheta,
                                  Eigen::VectorXf const &distsInPointsFrom,
                                  Eigen::MatrixX2f const &pointsTo,
                                  size_t const iterations, float const rho,
                                  float const epsilon) {

  Eigen::Vector3f theta = initTheta;

  Eigen::Vector3f meanSqGrad = Eigen::Vector3f::Zero(initTheta.rows());
  Eigen::Vector3f meanSqDTheta = Eigen::Vector3f::Zero(initTheta.rows());

  auto n = std::min(distsInPointsFrom.size(), pointsTo.rows());

  for (size_t i = 0; i < iterations; ++i) {
    auto grad = calcGradPosEst(theta, distsInPointsFrom, pointsTo, n);
    updateWithAdadelta(theta, meanSqGrad, meanSqDTheta, grad, rho, epsilon);
  }

  return theta;
}

Eigen::Vector3f IncPCA::calcGradPosEst(Eigen::Vector3f const &theta,
                                       Eigen::VectorXf const &distsInPointsFrom,
                                       Eigen::MatrixX2f const &pointsTo,
                                       size_t const n) {
  /// sigmoid ver
  Eigen::RowVector2f pos;
  pos << theta(0), theta(1);
  auto z = theta(2); // zeta for sigmoid function
  auto g = 0.1f;     // gain for sigmoid function
  auto s = 1.0f / (1.0f + std::exp(-g * z));
  auto c = 0.1f; // capacity allows scaling betwen 1-c to 1+C
  auto b = 1.0f - c + 2.0f * c * s;
  auto c2gs1_s = 2.0f * c * g * s * (1.0f - s);

  Eigen::MatrixX2f dPos = (-1.0f * pointsTo).rowwise() + pos;
  Eigen::VectorXf bl2Norm = b * dPos.rowwise().norm();
  bl2Norm = bl2Norm.unaryExpr([](float v) { return v > 1e-6f ? v : 1e-6f; });
  Eigen::VectorXf dist2_bl2Norm_1 =
      2.0f * (distsInPointsFrom.array() / bl2Norm.array() - 1.0f);

  Eigen::ArrayXf gradXY =
      ((-b * dPos).array().colwise() * dist2_bl2Norm_1.array()).colwise().sum();
  float gradZ =
      ((-c2gs1_s * dist2_bl2Norm_1).array() * bl2Norm.array() * bl2Norm.array())
          .sum();

  Eigen::Vector3f result;
  result << gradXY(0), gradXY(1), gradZ;

  return result;
}

float IncPCA::calcUncertU(Eigen::Vector3f const &theta,
                          Eigen::VectorXf const &distsInPointsFrom,
                          Eigen::MatrixX2f const &pointsTo) {
  float result = 0.0f;
  float err = errorPosEst(theta, distsInPointsFrom, pointsTo);
  float sumSqDistsInPointsFrom = distsInPointsFrom.array().square().sum();

  if (sumSqDistsInPointsFrom > 0.0f) {
    result = std::sqrt(err / sumSqDistsInPointsFrom);
  }
  if (err > sumSqDistsInPointsFrom) {
    // sometimes \beta * s'_ui is larger than s_ui. If so, ucert should be 1.0f
    result = 1.0f;
  }

  return result;
}

float IncPCA::errorPosEst(Eigen::Vector3f const &theta,
                          Eigen::VectorXf const &distsInPointsFrom,
                          Eigen::MatrixX2f const &pointsTo) {
  /// sigmoid ver
  Eigen::RowVector2f pos;
  pos << theta(0), theta(1);
  auto z = theta(2); // zeta for sigmoid function
  auto g = 0.1f;     // gain for sigmoid function
  auto s = 1.0f / (1.0f + std::exp(-g * z));
  auto c = 0.1f; // capacity allows scaling betwen 1-c to 1+C
  auto b = 1.0f - c + 2.0f * c * s;

  Eigen::MatrixX2f dPos = (-1.0f * pointsTo).rowwise() + pos;
  Eigen::VectorXf bl2Norm = b * dPos.rowwise().norm();

  return ((distsInPointsFrom - bl2Norm).array().square()).sum();
}

std::tuple<float, float, float> IncPCA::updateUncertWeight(
    float currentGamma, float currentSqGrad, float currentSqDGamma,
    Eigen::MatrixXf const &sigma, std::vector<Eigen::MatrixXf> const &sprimes,
    Eigen::MatrixXf const &uncertUs, Eigen::VectorXf const &uncertVs) {
  // sigma: shape(m, n)
  // sprimes: D x shape(m, n)
  // uncrertUs: shape(D, m)
  // uncrertVs: shape(D)

  Eigen::Index m = sigma.rows();                 // # of new data points
  Eigen::Index n = sigma.cols();                 // # of exisiting data points
  Eigen::Index D = Eigen::Index(sprimes.size()); // # of full dimensions

  // check matrix sizes
  for (auto const &sprime : sprimes) {
    if (sprime.rows() != m || sprime.cols() != n) {
      std::cerr << "Sigma and one of Sprimes have different matrix sizes"
                << std::endl;
    }
  }
  if (uncertUs.rows() != D || uncertUs.cols() != m) {
    std::cerr << "uncertUs does not have D rows and m cols" << std::endl;
  }
  if (uncertVs.size() != D) {
    std::cerr << "uncertVs does not have D length" << std::endl;
  }

  Eigen::MatrixXf errs(D, m);
  for (Eigen::Index l = 0; l < D; ++l) {
    // mean absolute error
    Eigen::RowVectorXf mae(m);
    mae = (sigma - sprimes[l]).array().abs().rowwise().mean();
    errs.row(l) = mae;
  }

  Eigen::RowVectorXf errD = errs.row(D - 1);
  Eigen::RowVectorXf meanU = uncertUs.colwise().mean();
  // added abs in case error with D dims > error with l dims (l <= D)
  Eigen::RowVectorXf sumDiffErr =
      (errs.rowwise() - errD).array().abs().colwise().sum();
  float sumUncertV = uncertVs.sum();

  // to avoid zero div
  meanU = meanU.unaryExpr([](float v) { return v > 1e-6f ? v : 1e-6f; });
  sumUncertV = std::max(sumUncertV, 1e-6f);

  Eigen::RowVectorXf rho = errD.array() / meanU.array();
  Eigen::RowVectorXf phi = sumDiffErr / sumUncertV;

  // Altough using a different gamma for each dim is possible,
  // this implementation using average of rho and phi for simplification
  float meanRho = rho.mean();
  float meanPhi = phi.mean();

  if (meanRho + meanPhi > 0) {
    // update with adadelta
    float r = 0.95;
    float e = 1e-6;

    float grad = currentGamma - meanRho / (meanRho + meanPhi);
    currentSqGrad = r * currentSqGrad + (1.0f - r) * grad * grad;
    float dGamma =
        -grad * std::sqrt(currentSqDGamma + e) / std::sqrt(currentSqGrad + e);

    currentSqDGamma = r * currentSqDGamma + (1.0f - r) * dGamma * dGamma;

    currentGamma = currentGamma + dGamma;
  }

  return std::make_tuple(currentGamma, currentSqGrad, currentSqDGamma);
}
