/* References
----------
T. Fujiwara, J.-K. Chou, Shilpika, P. Xu, L. Ren, K.-L. Ma. An Incremental
Dimensionality Reduction Method for Visualizing Streaming Multidimensional
Data, TVCG, 2020
*/

#ifndef IncPCA_HPP
#define IncPCA_HPP

#include "Eigen/Core"
#include "Eigen/Dense"
#include "Eigen/LU"
#include <array>
#include <tuple>
#include <vector>

class IncPCA {
public:
  IncPCA(Eigen::Index const nComponents = 2,
         float const forgettingFactor = 1.0f);

  void initialize();
  void partialFit(Eigen::MatrixXf X);
  Eigen::MatrixXf transform(Eigen::MatrixXf const &X);
  Eigen::MatrixXf getComponents();
  Eigen::MatrixXf getLoadings();
  float getUncertV(unsigned int const nObtainedFeatures);

  Eigen::MatrixX2f static geomTrans2D(Eigen::MatrixX2f const &pointsFrom,
                                      Eigen::MatrixX2f const &pointsTo);

  std::pair<Eigen::RowVector2f, float> static posEst(
      Eigen::RowVector2f const &posInPointsFrom,
      Eigen::MatrixX2f const &pointsFrom, Eigen::MatrixX2f const &pointsTo);

  std::tuple<float, float, float> static updateUncertWeight(
      float currentGamma, float currentSqGrad, float currentSqDGamma,
      Eigen::MatrixXf const &sigma, std::vector<Eigen::MatrixXf> const &sprimes,
      Eigen::MatrixXf const &uncertUs, Eigen::VectorXf const &uncertVs);

private:
  Eigen::Index nComponents;
  float forgettingFactor;

  Eigen::MatrixXf components_;
  Eigen::Index nComponents_;
  Eigen::Index nSamplesSeen_;
  Eigen::ArrayXf mean_;
  Eigen::ArrayXf var_;
  Eigen::MatrixXf singularValues_;
  Eigen::MatrixXf explainedVariance_;
  Eigen::MatrixXf explainedVarianceRatio_;
  float noiseVariance_;

  std::tuple<
      Eigen::ArrayXf, Eigen::ArrayXf,
      Eigen::Index> static incrementalMeanAndVar(Eigen::MatrixXf const &X,
                                                 float forgettingFactor,
                                                 Eigen::ArrayXf const
                                                     &lastMean = {},
                                                 Eigen::ArrayXf const
                                                     &lastVariance = {},
                                                 Eigen::Index const
                                                     lastSampleCount = 0);
  void static svdFlip(Eigen::MatrixXf &U, Eigen::MatrixXf &V);

  Eigen::Vector3f static optPosEst(Eigen::Vector3f const &initTheta,
                                   Eigen::VectorXf const &distsInPointsFrom,
                                   Eigen::MatrixX2f const &pointsTo,
                                   size_t const iterations = 1000,
                                   float const rho = 0.95,
                                   float const epsilon = 1e-6);
  Eigen::Vector3f static calcGradPosEst(
      Eigen::Vector3f const &theta, Eigen::VectorXf const &distsInPointsFrom,
      Eigen::MatrixX2f const &pointsTo, size_t const n);
  void static updateWithAdadelta(Eigen::Vector3f &theta,
                                 Eigen::Vector3f &meanSqGrad,
                                 Eigen::Vector3f &meanSqDTheta,
                                 Eigen::Vector3f const &grad, float const rho,
                                 float const epsilon);
  float static calcUncertU(Eigen::Vector3f const &theta,
                           Eigen::VectorXf const &distsInPointsFrom,
                           Eigen::MatrixX2f const &pointsTo);
  float static errorPosEst(Eigen::Vector3f const &theta,
                           Eigen::VectorXf const &distsInPointsFrom,
                           Eigen::MatrixX2f const &pointsTo);
};

#endif // IncPCA_HPP
