/* References
----------
T. Fujiwara, J.-K. Chou, Shilpika, P. Xu, L. Ren, K.-L. Ma. An Incremental
Dimensionality Reduction Method for Visualizing Streaming Multidimensional
Data, TVCG, 2020
*/

#include "inc_pca.hpp"

#include <chrono>
#include <iostream>
#include <mutex>
#include <thread>

std::mutex mtx;

int main(int argc, char *argv[]) {
  Eigen::Index nMeasures = 10;
  Eigen::Index m = 2;
  size_t m_ = size_t(m);
  size_t numWorkers = m_;

  std::vector<Eigen::Index> ds = {10, 100, 1000};
  std::vector<Eigen::Index> ns = {100, 1000, 10000};

  auto start = std::chrono::steady_clock::now();
  auto end = std::chrono::steady_clock::now();
  auto duration =
      std::chrono::duration_cast<std::chrono::microseconds>(end - start)
          .count();

  for (auto const &d : ds) {
    for (auto const &n : ns) {
      std::cout << "========================" << std::endl << std::endl;
      std::cout << "n: " << n << ", d: " << d << std::endl;

      unsigned long ipcaMicroSec = 0;
      unsigned long geomTransMicroSec = 0;
      unsigned long updateBetaMicroSec = 0;
      unsigned long posEstMicroSec = 0;
      unsigned long uncertVMicroSec = 0;

      Eigen::MatrixXf X;
      X.setRandom(n + m, d);
      Eigen::MatrixXf X1 = X.topRows(n);
      Eigen::MatrixXf X2 = X.bottomRows(m);
      Eigen::MatrixXf X1MissingLast = X1.leftCols(d - 1);
      Eigen::MatrixXf X2MissingLast = X2.leftCols(d - 1);

      // 1. inc pca
      // preparation (loading n exiting points)
      IncPCA ipca(2, 1.0f);
      std::cout << "computing ipca for " << n << " exiting points" << std::endl;
      ipca.partialFit(X1);
      Eigen::MatrixX2f Y1 = ipca.transform(X1);
      std::cout << "computation done for exiting points" << std::endl;
      // preparation done

      Eigen::MatrixX2f Y;
      for (Eigen::Index count = 0; count < nMeasures; ++count) {
        start = std::chrono::steady_clock::now();

        ipca.partialFit(X2);
        Y = ipca.transform(X);

        end = std::chrono::steady_clock::now();
        duration =
            std::chrono::duration_cast<std::chrono::microseconds>(end - start)
                .count();
        ipcaMicroSec += duration;
      }
      ipcaMicroSec /= nMeasures;
      std::cout << "ipca: " << ipcaMicroSec / 1000.0 << " millisec" << std::endl;

      // 2. geom transform
      for (Eigen::Index count = 0; count < nMeasures; ++count) {
        start = std::chrono::steady_clock::now();

        Eigen::MatrixX2f Y_geom = IncPCA::geomTrans2D(Y1, Y);

        end = std::chrono::steady_clock::now();
        duration =
            std::chrono::duration_cast<std::chrono::microseconds>(end - start)
                .count();
        geomTransMicroSec += duration;
      }
      geomTransMicroSec /= nMeasures;
      std::cout << "geom: " << geomTransMicroSec / 1000.0 << " millisec" << std::endl;

      // 3. update beta
      float beta = 0.1f;
      float sqGrad = 0.1f;
      float sqDbeta = 0.1f;

      std::vector<Eigen::MatrixXf> sprimes(d, Eigen::MatrixXf::Random(m, n));
      Eigen::MatrixXf sigma = Eigen::MatrixXf::Random(m, n);
      Eigen::VectorXf uncertVs = Eigen::VectorXf::Random(d);
      Eigen::MatrixXf uncertUs = Eigen::MatrixXf::Random(d, m);

      for (Eigen::Index count = 0; count < nMeasures; ++count) {
        start = std::chrono::steady_clock::now();

        IncPCA::updateUncertWeight(beta, sqGrad, sqDbeta, sigma, sprimes,
                                   uncertUs, uncertVs);

        end = std::chrono::steady_clock::now();
        duration =
            std::chrono::duration_cast<std::chrono::microseconds>(end - start)
                .count();
        updateBetaMicroSec += duration;
      }
      updateBetaMicroSec /= nMeasures;
      std::cout << "beta: " << updateBetaMicroSec / 1000.0 << " millisec" << std::endl;

      // 4. pos estimation
      // preparation
      IncPCA ipcaMissingLast(2, 1.0f);
      ipcaMissingLast.partialFit(X1MissingLast);
      Eigen::MatrixX2f Y1MissingLast = ipcaMissingLast.transform(X1MissingLast);
      // preparation done

      for (Eigen::Index count = 0; count < nMeasures; ++count) {
        std::vector<std::thread> worker;

        start = std::chrono::steady_clock::now();

        Eigen::MatrixX2f positionsMissingLast =
            ipcaMissingLast.transform(X2MissingLast);

        for (size_t w = 0; w < numWorkers; ++w) {
          worker.emplace_back(
              [&](size_t id) {
                auto r0 = m_ / numWorkers * id + std::min(m_ % numWorkers, id);
                auto r1 = m_ / numWorkers * (id + 1) +
                          std::min(m_ % numWorkers, id + 1);
                for (auto i = r0; i < r1; ++i) {
                  IncPCA::posEst(positionsMissingLast.row(i), Y1MissingLast,
                                 Y1);
                }
              },
              w);
        }
        for (auto &w : worker)
          w.join();

        end = std::chrono::steady_clock::now();
        duration =
            std::chrono::duration_cast<std::chrono::microseconds>(end - start)
                .count();
        posEstMicroSec += duration;
      }

      posEstMicroSec /= nMeasures;
      std::cout << "posEst: " << posEstMicroSec / 1000.0 << " millisec" << std::endl;

      // 5. update uncert V
      for (Eigen::Index count = 0; count < nMeasures; ++count) {
        start = std::chrono::steady_clock::now();
        for (Eigen::Index i = 0; i < d; ++i) {
          uncertVs[i] = ipca.getUncertV(i);
        }
        end = std::chrono::steady_clock::now();
        duration =
            std::chrono::duration_cast<std::chrono::microseconds>(end - start)
                .count();
        uncertVMicroSec += duration;
      }
      uncertVMicroSec /= nMeasures;

      std::cout << "uncertV: " << uncertVMicroSec / 1000.0 << " millisec" << std::endl;
    }
  }
  return 0;
}
